#include "lcd_modul.h"
#include <wiringPi.h>
#include <iostream>

using namespace std;

LCD::LCD()
{
    if (wiringPiSetup() == -1) {
        cerr << "wiringPiSetup() failed\n";
        exit(1);
    }
	/*
	lcdInit (const int rows, const int cols, const int bits,
	const int rs, const int strb,
	const int d0, const int d1, const int d2, const int d3, const int d4,
	const int d5, const int d6, const int d7)
	*/
    fd = lcdInit(2, 16, 4, 2, 3, 6, 5, 4, 1, 0, 0, 0, 0);
    if (fd == -1) {
        cerr << "lcdInit() failed\n";
        exit(1);
    }

    //Definiere eigene Zeichen
    u_char copyright[] = {0b01110, 0b10001, 0b11101, 0b11001,
                          0b11001, 0b11101, 0b10001, 0b01110};
    lcdCharDef(fd, COPYRIGHT, copyright);

    u_char szlig[] = {0b01110, 0b10001, 0b10001, 0b11110,
                      0b10001, 0b10001, 0b11110, 0b10000};
    lcdCharDef(fd, SZLIG, szlig);

    u_char smile[] = {0b01110, 0b10001, 0b11011, 0b10001,
                      0b11011, 0b10101, 0b10001, 0b01110};
    lcdCharDef(fd, SMILE, smile);

    u_char sad[] = {0b01110, 0b10001, 0b11011, 0b10001,
                    0b10101, 0b11011, 0b10001, 0b01110};
    lcdCharDef(fd, SAD, sad);

    u_char heart[] = {0b00000, 0b01010, 0b10101, 0b10001,
                      0b01010, 0b01010, 0b00100, 0b00000};
    lcdCharDef(fd, HEART, heart);

    u_char upperAWithDots[] = {0b01010, 0b00000, 0b01110, 0b10001,
                               0b11111, 0b10001, 0b10001, 0b00000};
    lcdCharDef(fd, UPPER_A_UMLAUT, upperAWithDots);

    clear();
    display(0, "LCD Steuerung");
    display(1, COPYRIGHT + string {"hulemule.de"});
}

LCD::~LCD()
{
    for (int i {0}; i != 2; ++i) {
        stopDisplaying[i] = true;
        unique_lock<mutex> lockLine {blockLine[i]};
        mode[i] = Mode::fixed;
    }
    clear();

    display(0, string {"Und tsch"} + U_UMLAUT + "ss,");
    display(1, string {"bis demn"} + LOWER_A_UMLAUT + "chst!");
    delay(5000);
    clear();
}

void LCD::clear()
{
    lcdClear(fd);
}

void LCD::displayFixedText(int line, const string& text)
{
    unique_lock<mutex> lock {uniqueAccessToLcdWrite};
    lcdPosition(fd, 0, line);
    lcdPuts(fd, text.c_str());
}

void LCD::displayChar(int line, int pos, const u_char c)
{
    lcdPosition(fd, pos, line);
    lcdPutchar(fd, c);
}

void LCD::displayFlowingText(int line, const string& text)
{
    u_int pos {0};
    unique_lock<mutex> lockLine {blockLine[line]};
    while (!stopDisplaying[line]) {
        {
            unique_lock<mutex> lockWrite {uniqueAccessToLcdWrite};
            for (u_int i {pos}; i != pos + 16; ++i) {
                displayChar(line, i - pos,
                            i < 16 || i > 15 + text.size() ? ' ' : text[i - 16]);
            }
        }
        delay(500 - 4*velocity);
        ++pos;
        pos %= (16 + text.size());
    }
}

static const char blankLine[] {"                "};

void LCD::displayEffectText(int line, const string& text)
{
    unique_lock<mutex> lockLine {blockLine[line]};
    while (true) {
        for (u_int i {0}; i != text.size(); ++i) {
            for (u_int j {15}; j != i - 1; --j) {
                {
                    unique_lock<mutex> lockWrite {uniqueAccessToLcdWrite};
                    if (j != 15)
                        displayChar(line, j + 1, ' ');
                    displayChar(line, j, text[i]);
                }
                if (stopDisplaying[line]) {
                    displayFixedText(line, blankLine);
                    return;
                }
                if (text[i] != ' ')
                    delay(300 - 2.5*velocity);
            }
            if (i == text.size() -1)
                delay(1000 - 9*velocity);
        }
        displayFixedText(line, blankLine);
    }
}

void LCD::displayBlinkingText(int line, const string& text)
{
    unique_lock<mutex> lockLine {blockLine[line]};
    bool showText {true};
    while (!stopDisplaying[line]) {
        displayFixedText(line, showText ? text : blankLine);
        delay(2000 - 19*velocity);
        showText = !showText;
    }
    displayFixedText(line, blankLine);
}

void LCD::displayTypedText(const string& text1, const string& text2)
{
    unique_lock<mutex> lockLine1 {blockLine[0]};
    unique_lock<mutex> lockLine2 {blockLine[1]};
    while (true) {
        lcdCursor(fd, 1);
        for (int i {0}; i != 2; ++i) {
            const string& text = i == 0 ? text1: text2;
            for (u_int j {0}; j != text.size(); ++j) {
                displayChar(i, j, text[j]);
                if (stopDisplaying[0] || stopDisplaying[1])
                    break;
                if (i == 1 && j == 15)
                    lcdCursor(fd, 0);
                delay(300 - 2.5*velocity);
            }
        }
        if (stopDisplaying[0] || stopDisplaying[1])
            break;
        delay(1000 - 8*velocity);
        clear();
    }
    clear();
    lcdCursor(fd, 0);
}

void LCD::display(int line, const string& text, const string& text2)
{
    stopDisplaying[line] = true;
    unique_lock<mutex> lockLine {blockLine[line]};
    displayFixedText(line, blankLine);
    stopDisplaying[line] = false;
    if (mode[line] == Mode::fixed) {
        displayFixedText(line, text);
    }
    else if (mode[line] == Mode::flowing) {
        thread t{&LCD::displayFlowingText, this, line, text};
        t.detach();
    }
    else if (mode[line] == Mode::effect) {
        thread t{&LCD::displayEffectText, this, line, text};
        t.detach();
    }
    else if (mode[line] == Mode::blinking) {
        thread t{&LCD::displayBlinkingText, this, line, text};
        t.detach();
    }
    else if (mode[line] == Mode::typed) {
        stopDisplaying[1] = true;
        unique_lock<mutex> lockLine2 {blockLine[1]};
        displayFixedText(1, blankLine);
        stopDisplaying[1] = false;
        thread t{&LCD::displayTypedText, this, text, text2};
        t.detach();
    }
}
